import sys
import re
import json

from collections import defaultdict
from collections.abc import Mapping
from fhirstore.search import SearchArguments, ReverseChain

number_prefix_matching = {"gt": "gt", "ge": "gte", "lt": "lt", "le": "lte"}


def build_element_query(key, value):
    """Translate and parse search parameters (key, value) to an
    elasticSearch element query
    """
    element_query = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

    if isinstance(value, int) or isinstance(value, float):
        element_query["match"][key] = value

    else:
        numeric_prefix = re.search(r"^(gt|lt|ge|le)([0-9].*)$", value)
        eq_prefix = re.search(r"^(eq)([0-9].*)$", value)
        special_prefix = re.search(r"^(ne|sa|eb|ap)([0-9].*)$", value)
        pipe_suffix = re.search(r"(.*)\|(.*)", value)

        string_modif = re.search(
            r"^(.*):(contains|exact|above|below|not|in|not-in|of-type|identifier)$", key
        )
        if key == "_text" or key == "_content":
            element_query["simple_query_string"]["query"] = value

        elif string_modif:
            string_modifier = string_modif.group(2)
            string_field = string_modif.group(1)
            if string_modifier == "contains":
                element_query["query_string"]["query"] = f"*{value}*"
                element_query["query_string"]["fields"] = [string_field]

            elif string_modifier == "exact":
                element_query["simple_query_string"]["query"] = f'"{value}"'
                element_query["simple_query_string"]["fields"] = [string_field]
                element_query["simple_query_string"]["flags"] = "PHRASE"

            elif string_modifier == "not":
                element_query["simple_query_string"]["query"] = f"-{value}"
                element_query["simple_query_string"]["fields"] = [string_field]

            elif string_modifier == "not-in":
                element_query["simple_query_string"]["query"] = f"-{value}"
                element_query["simple_query_string"]["fields"] = [string_field]

            elif string_modifier == "in":
                element_query["simple_query_string"]["query"] = f'"{value}"'
                element_query["simple_query_string"]["fields"] = [string_field]
                element_query["simple_query_string"]["flags"] = "PHRASE"

            elif string_modifier == "below":
                element_query["simple_query_string"]["query"] = f"({value})*"
                element_query["simple_query_string"]["fields"] = [string_field]

            elif string_modifier == "identifier":
                element_query["simple_query_string"]["query"] = f'"{value}"'
                element_query["simple_query_string"]["fields"] = [
                    f"{string_field}.identifier.value"
                ]
                element_query["simple_query_string"]["flags"] = "PHRASE"

        elif numeric_prefix:
            element_query["range"][key] = {
                number_prefix_matching[numeric_prefix.group(1)]: numeric_prefix.group(2)
            }
        elif eq_prefix:
            element_query["match"][key] = eq_prefix.group(2)

        elif special_prefix:
            if special_prefix.group(1) == "ne":
                element_query["simple_query_string"]["query"] = f"-{special_prefix.group(2)}"
                element_query["simple_query_string"]["fields"] = [key]

        elif pipe_suffix:
            system, code = re.split(r"\|", value)
            element_query["bool"]["must"] = [
                {
                    "simple_query_string": {
                        "query": f'"{system}"',
                        "fields": [f"{key}.system"],
                        "flags": "PHRASE",
                    }
                }
            ]
            element_query["bool"]["must"].append(
                {
                    "simple_query_string": {
                        "query": f'"{code}"',
                        "fields": [f"{key}.code", f"{key}.value"],
                        "flags": "PHRASE",
                    }
                }
            )

        elif isinstance(value, str):
            element_query["simple_query_string"]["query"] = f'"{value}"'
            element_query["simple_query_string"]["fields"] = [key]
            element_query["simple_query_string"]["flags"] = "PHRASE"

    return element_query


def validate_parameters(core_args):
    """Validates that parameters is in dictionary form
    """
    assert isinstance(core_args, Mapping), "parameters must be a dictionary"


def build_simple_query(dict_args):
    """Translates a dictionary of length 1 to
    a simple elasticsearch query
    """
    validate_parameters(dict_args)

    if dict_args.get("multiple"):
        multiple_key = list(dict_args["multiple"])[0]
        multiple_values = dict_args["multiple"][multiple_key]
        content = [build_element_query(multiple_key, element) for element in multiple_values]
        sub_query = {"bool": {"should": content}}
    else:
        key = list(dict_args)[0]
        values = dict_args.get(key)
        if len(values) == 1:
            sub_query = build_element_query(key, values[0])
        else:
            content = [build_element_query(key, element) for element in values]
            sub_query = {"bool": {"must": content}}
    return sub_query


def build_core_query(core_args):
    """Translates a full JSON body to
    the core of an elasticsearch query
    """
    validate_parameters(core_args)

    if core_args == {}:
        query = {"match_all": {}}
    elif len(core_args) == 1:
        query = build_simple_query(core_args)
    elif len(core_args) > 1:
        inter_query = [build_simple_query({sub_key: core_args[sub_key]}) for sub_key in core_args]
        query = {"bool": {"must": inter_query}}
    return query



import elasticsearch

from fhirpath.connectors import create_connection
from fhirpath.engine.es import ElasticsearchEngine as BaseEngine
from fhirpath.engine import dialect_factory
from fhirpath.enums import FHIR_VERSION, EngineQueryType
from fhirpath_helpers.elasticsearch.mapping import generate_mappings


class ElasticSearchEngine(BaseEngine):
    def __init__(self, fhir_release, conn_factory, dialect_factory):
        super().__init__(fhir_release, conn_factory, dialect_factory)
        self.mappings = generate_mappings(FHIR_VERSION.R4.name)

    def calculate_field_index_name(self, resource_type):
        return resource_type

    def current_url(self):
        """
        complete url from current request
        return yarl.URL"""
        from yarl import URL

        return URL("https://dev.arkhn.com")

    def get_index_name(self):
        """ """
        # return f"fhirstore.{resource_type.lower()}"
        return "fhirstore"

    def get_mapping(self, resource_type: str, index_config=None):
        """ """
        try:
            return {"properties": self.mappings[resource_type]}
        except KeyError as e:
            raise Exception(f"resource_type {e} does not exist in elasticsearch mappings")

    def _execute(self, query, unrestricted, query_type):
        """ """
        # for now we support single from resource
        query_copy = query.clone()
        resource_type = query.get_from()[0][1].resource_type
        field_index_name = self.calculate_field_index_name(resource_type)

        if unrestricted is False:
            self.build_security_query(query_copy)

        params = {
            "query": query_copy,
            "root_replacer": field_index_name,
            "mapping": self.get_mapping(resource_type),
        }

        compiled = self.dialect.compile(**params)
        if query_type == EngineQueryType.DML:
            raw_result = self.connection.fetch(self.get_index_name(resource_type), compiled)
        elif query_type == EngineQueryType.COUNT:
            raw_result = self.connection.count(self.get_index_name(resource_type), compiled)
        else:
            raise NotImplementedError

        return raw_result, field_index_name, compiled

    def create_es_index(self):
        body = {
            "settings": {
                "index.mapping.total_fields.limit": 10000,
                "index.mapping.nested_fields.limit": 10000,
                "analysis": {
                    "analyzer": {
                        "path_analyzer": {"tokenizer": "path_tokenizer"},
                        "fhir_reference_analyzer": {"tokenizer": "fhir_reference_tokenizer"},
                    },
                    "char_filter": {},
                    "filter": {},
                    "tokenizer": {
                        "path_tokenizer": {"delimiter": "/", "type": "path_hierarchy"},
                        "fhir_reference_tokenizer": {"type": "pattern", "pattern": "/"},
                    },
                },
            },
            "mappings": {
                "dynamic": False,
                "properties": {
                    "elastic_index": {"index": True, "store": True, "type": "keyword"},
                    "id": {"index": True, "store": True, "type": "keyword"},
                },
            },
        }

        for resource_type in self.mappings.keys():
            body["mappings"]["properties"][resource_type] = self.get_mapping(resource_type)

        self.connection._conn.indices.create(self.get_index_name(), body=body)
        self.connection._conn.indices.refresh(index=self.get_index_name())


def create_search_engine(es_client: elasticsearch.Elasticsearch):
    es_node = es_client.transport.hosts[0]
    connection = create_connection(
        f"es://{es_node.get('http_auth', 'elastic')}@{es_node['host']}:{es_node['port']}/",
        "elasticsearch.Elasticsearch",
    )
    return ElasticSearchEngine(FHIR_VERSION.R4, lambda x: connection, dialect_factory)



import pytest
import json
from pytest import raises
from fhirstore.search import SearchArguments
from fhirstore.search import Bundle
from collections.abc import Mapping
from fhirstore import FHIRStore, NotFoundError
from werkzeug.datastructures import ImmutableMultiDict


def test_fill_search_bundle():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([]), "Patient")
    bundle = Bundle()
    bundle.fill({}, search_args.formatting_args)
    assert bundle.content == {
        "resource_type": "Bundle",
        "entry": [],
        "total": 0,
    }


def test_fill_count_bundle():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_summary", "count")]), "Patient")
    bundle = Bundle()
    bundle.fill({}, search_args.formatting_args)
    assert bundle.content == {
        "resource_type": "Bundle",
        "tag": {"code": "SUBSETTED"},
        "total": 0,
    }


def test_fill_tag_search_bundle():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_summary", "text")]), "Patient")
    bundle = Bundle()
    bundle.fill({}, search_args.formatting_args)
    assert bundle.content == {
        "resource_type": "Bundle",
        "entry": [],
        "tag": {"code": "SUBSETTED"},
        "total": 0,
    }


def test_fill_no_tag_search_bundle():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_summary", "false")]), "Patient")
    bundle = Bundle()
    bundle.fill({}, search_args.formatting_args)
    assert bundle.content == {
        "resource_type": "Bundle",
        "entry": [],
        "total": 0,
    }


def test_fill_bundle_search():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([]), "Patient")
    bundle = Bundle()
    bundle.fill(
        {
            "took": 2,
            "timed_out": False,
            "_shards": {"total": 1, "successful": 1, "skipped": 0, "failed": 0},
            "hits": {
                "total": {"value": 1, "relation": "eq"},
                "max_score": 1.0925692,
                "hits": [
                    {
                        "_index": "fhirstore.patient",
                        "_type": "_doc",
                        "_id": "5e96c51977d71dc87644653b",
                        "_score": 1.0925692,
                        "_source": {
                            "active": True,
                            "contact": [
                                {
                                    "organization": {
                                        "display": "Walt Disney Corporation",
                                        "reference": "Organization/1",
                                    }
                                }
                            ],
                        },
                    }
                ],
            },
        },
        search_args.formatting_args,
    )
    assert bundle.content == {
        "resource_type": "Bundle",
        "entry": [
            {
                "resource": {
                    "active": True,
                    "contact": [
                        {
                            "organization": {
                                "display": "Walt Disney Corporation",
                                "reference": "Organization/1",
                            }
                        }
                    ],
                },
                "search": {"mode": "match"},
            }
        ],
        "total": 1,
    }


def test_complete_bundle():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([]), "Patient")
    bundle = Bundle()
    bundle.fill(
        {
            "took": 2,
            "timed_out": False,
            "_shards": {"total": 1, "successful": 1, "skipped": 0, "failed": 0},
            "hits": {
                "total": {"value": 1, "relation": "eq"},
                "max_score": 1.0925692,
                "hits": [
                    {
                        "_index": "fhirstore.patient",
                        "_type": "_doc",
                        "_id": "5e96c51977d71dc87644653b",
                        "_score": 1.0925692,
                        "_source": {
                            "active": True,
                            "contact": [
                                {
                                    "organization": {
                                        "display": "Walt Disney Corporation",
                                        "reference": "Organization/1",
                                    }
                                }
                            ],
                        },
                    }
                ],
            },
        },
        search_args.formatting_args,
    )
    new_bundle = Bundle()
    new_bundle.content = {
        "entry": [
            {
                "resource": {
                    "active": True,
                    "birthDate": "1932-09-24",
                    "gender": "male",
                    "id": "xcda",
                    "name": [{"family": "Levin", "given": ["Henry"]}],
                    "resourceType": "Patient",
                },
                "search": {"mode": "match"},
            }
        ],
        "resource_type": "Bundle",
        "total": 1,
    }
    bundle.complete(new_bundle, search_args.formatting_args)
    assert bundle.content["total"] == 2
    assert bundle.content["entry"] == [
        {
            "resource": {
                "active": True,
                "contact": [
                    {
                        "organization": {
                            "display": "Walt Disney Corporation",
                            "reference": "Organization/1",
                        }
                    }
                ],
            },
            "search": {"mode": "match"},
        },
        {
            "resource": {
                "active": True,
                "birthDate": "1932-09-24",
                "gender": "male",
                "id": "xcda",
                "name": [{"family": "Levin", "given": ["Henry"]}],
                "resourceType": "Patient",
            },
            "search": {"mode": "match"},
        },
    ]


def test_complete_bundle_count():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_summary", "count")]), "Patient")
    bundle = Bundle()
    bundle.fill(
        {"count": 1, "_shards": {"total": 1, "successful": 1, "skipped": 0, "failed": 0}},
        search_args.formatting_args,
    )
    other_bundle = Bundle()
    other_bundle.content = {"resource_type": "Bundle", "tag": {"code": "SUBSETTED"}, "total": 1}
    bundle.complete(other_bundle, search_args.formatting_args)
    assert bundle.content["total"] == 2
    assert bundle.content == {
        "resource_type": "Bundle",
        "tag": {"code": "SUBSETTED"},
        "total": 2,
    }


def test_add_included_bundle_no_include():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([]), "Patient")
    bundle = Bundle()
    bundle.append(
        {
            "took": 2,
            "timed_out": False,
            "_shards": {"total": 1, "successful": 1, "skipped": 0, "failed": 0},
            "hits": {
                "total": {"value": 1, "relation": "eq"},
                "max_score": 1.0925692,
                "hits": [
                    {
                        "_index": "fhirstore.patient",
                        "_type": "_doc",
                        "_id": "5e96c51977d71dc87644653b",
                        "_score": 1.0925692,
                        "_source": {
                            "active": True,
                            "contact": [
                                {
                                    "organization": {
                                        "display": "Walt Disney Corporation",
                                        "reference": "Organization/1",
                                    }
                                }
                            ],
                        },
                    }
                ],
            },
        },
        search_args.formatting_args,
    )
    assert bundle.content == {
        "resource_type": "Bundle",
        "entry": [],
        "total": 0,
    }


def test_add_included_bundle():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_include", "MedicationRequest:subject")]), "Patient")
    bundle = Bundle()
    bundle.append(
        {
            "took": 2,
            "timed_out": False,
            "_shards": {"total": 1, "successful": 1, "skipped": 0, "failed": 0},
            "hits": {
                "total": {"value": 1, "relation": "eq"},
                "max_score": 1.0925692,
                "hits": [
                    {
                        "_index": "fhirstore.patient",
                        "_type": "_doc",
                        "_id": "5e96c51977d71dc87644653b",
                        "_score": 1.0925692,
                        "_source": {
                            "active": True,
                            "contact": [
                                {
                                    "organization": {
                                        "display": "Walt Disney Corporation",
                                        "reference": "Organization/1",
                                    }
                                }
                            ],
                        },
                    }
                ],
            },
        },
        search_args.formatting_args,
    )
    assert bundle.content == {
        "resource_type": "Bundle",
        "entry": [
            {
                "resource": {
                    "active": True,
                    "contact": [
                        {
                            "organization": {
                                "display": "Walt Disney Corporation",
                                "reference": "Organization/1",
                            }
                        }
                    ],
                },
                "search": {"mode": "include"},
            }
        ],
        "total": 0,
    }



ARKHN_TERMINOLOGY_SYSTEM = "http://terminology.arkhn.org/CodeSystem"


class ARKHN_CODE_SYSTEMS:
    resource = f"{ARKHN_TERMINOLOGY_SYSTEM}/resource"
    source = f"{ARKHN_TERMINOLOGY_SYSTEM}/source"


from .fhirstore import FHIRStore, BadRequestError, NotFoundError  # noqa


from .searcharguments import SearchArguments, ReverseChain, parse_comma, pre_process_params
from .corequerybuilder import build_core_query, build_simple_query, build_element_query
from .bundle import Bundle



from collections import defaultdict
from collections.abc import Mapping
from werkzeug.datastructures import ImmutableMultiDict

import re


def url_to_dict(url_args):
    if isinstance(url_args, ImmutableMultiDict):
        search_args = {key: url_args.getlist(key) for key in url_args.keys()}
        return search_args
    else:
        return url_args


def parse_comma(key, value):
    if "," in value:
        return "multiple", {key: value.split(",")}
    else:
        return key, [value]


def pre_process_params(url_args):
    search_args = url_to_dict(url_args)
    processed_params = defaultdict(list)
    if search_args == {}:
        processed_params = {}
    for key, value in search_args.items():
        if len(value) == 1:
            parsed_key, parsed_value = parse_comma(key, value[0])
            processed_params.update({parsed_key: parsed_value})
        else:
            for element in value:
                parsed_key, parsed_value = parse_comma(key, element)
                if parsed_key == "multiple":
                    processed_params[parsed_key] = parsed_value
                else:
                    processed_params[parsed_key].append(parsed_value[0])
    return processed_params


class ReverseChain:
    """Reverse Chaining is used to select resources based on the properties of
    resources that refer to them.
    Example: Patient?_has:Observation:patient:code=1234-5 returns Patient resources,
    where the patient resource is referred to by at least one Observation where the
    observation has a code of 1234, and where the Observation refers to the patient
    resource in the patient search parameter
    """

    def __init__(self):
        self.is_queried = False
        self.resources_type = []
        self.references = []
        self.fields = []

    def parse(self, key, value):
        self.is_queried = key.startswith("_has:")
        if self.is_queried == True:
            self.value = value
            chain_elements = key.split(":")
            self.resources_type.append(chain_elements[1])
            self.references.append(chain_elements[2])
            if chain_elements[3] == "_has":
                self.resources_type.append(chain_elements[4])
                self.references.append(chain_elements[5])
                self.fields.append(chain_elements[6])
            else:
                self.fields.append(chain_elements[3])

    def format(self):
        if self.is_queried == True:
            self.has_args = []
            if len(self.resources_type) == 1:
                self.has_args.append(
                    {"_elements": [f"{self.references[0]}.reference"], self.fields[0]: []}
                )
            else:
                self.has_args.append(
                    {"_elements": [f"{self.references[1]}.reference"], self.fields[1]: self.value}
                )
                self.has_args.append(
                    {"_elements": [f"{self.references[0]}.reference"], self.fields[0]: []}
                )


class SearchArguments:
    def __init__(self):
        self.core_args = {}
        self.formatting_args = {
            "include": None,
            "is_summary_count": False,
            "summary": False,
            "elements": None,
            "sort": None,
        }
        self.meta_args = {"offset": 0, "result_size": 100}

    def sort_params(self, args):
        has_sort = None
        if "_sort" in args:
            has_sort = args["_sort"]
        elif "multiple" in args:
            has_sort = args["multiple"].get("_sort", None)
        # if there is a sorting argument, process it to handle a change of sorting order
        if has_sort:
            sorting_params = []
            for argument in has_sort:
                # find a "-" before the argument. It indicates a sorting order different from default
                has_minus = argument.startswith("-")
                if has_minus:
                    # if the argument is -score, sort by ascending order (i.e. ascending relevance)
                    if argument[1:] == "_score":
                        sorting_params.append({argument[1:]: {"order": "asc"}})
                    # for any other argument, sort by descending order
                    else:
                        sorting_params.append({argument[1:]: {"order": "desc"}})
                # if there is no "-", use order defaults defined in elasticsearch
                else:
                    sorting_params.append(argument)
            return sorting_params

    def include_params(self, args):
        include = None
        if "_include" in args:
            include = [re.search(r"(.*):(.*)", args["_include"][0]).group(2)]

        elif "multiple" in args and "_include" in args["multiple"]:
            attributes = args["multiple"].get("_include", None)
            include = [re.search(r"(.*):(.*)", elem).group(2) for elem in attributes]
        return include

    def has_reverse_chain(self, args):
        reverse_chain = ReverseChain()
        # TODO handle independent _has
        for key in list(args):
            if key.startswith("_has:"):
                reverse_chain.parse(key, args[key])
                args.pop(key)
                reverse_chain.format()
        return reverse_chain

    def clean_params(self, args):
        if "multiple" in args:
            args["multiple"].pop("_elements", None)
            args["multiple"].pop("_sort", None)
            args["multiple"].pop("_include", None)
            args["multiple"].pop("_type", None)

        if args.get("multiple") == {}:
            args.pop("multiple")

        args.pop("_summary", None)
        args.pop("_elements", None)
        args.pop("_sort", None)
        args.pop("_include", None)
        args.pop("_count", None)

        return args

    def parse(self, url_args, resource_type):
        self.resource_type = resource_type
        args = pre_process_params(url_args)
        self.formatting_args["sort"] = self.sort_params(args)
        self.reverse_chain = self.has_reverse_chain(args)
        self.formatting_args["include"] = self.include_params(args)
        has_result_size = args.get("_count", None)
        self.meta_args["result_size"] = int(has_result_size[0]) if has_result_size else 100
        self.formatting_args["summary"] = "_summary" in args and args["_summary"] != ["false"]
        self.formatting_args["is_summary_count"] = (
            "_summary" in args and args["_summary"][0] == "count"
        )

        if "_summary" in args and args["_summary"][0] == "text":
            self.formatting_args["elements"] = ["text", "id", "meta"]
        elif "_elements" in args:
            self.formatting_args["elements"] = args["_elements"]
        elif "multiple" in args:
            self.formatting_args["elements"] = args["multiple"].get("_elements", None)

        self.core_args = self.clean_params(args)



import pytest
import json
from pytest import raises

from time import sleep
from fhirstore import FHIRStore, NotFoundError
from collections.abc import Mapping
from werkzeug.datastructures import ImmutableMultiDict

# For now, this class assumes an already existing store exists
# (store.bootstrap was run)


@pytest.fixture(scope="module")
def insert_patient(es_client):
    if not es_client.indices.exists("fhirstore.patient"):
        with open("test/fixtures/patient-example.json") as f:
            patient_1 = json.load(f)
            es_client.index(index="fhirstore.patient", body=patient_1)

        with open("test/fixtures/patient-example-2.json") as g:
            patient_2 = json.load(g)
            es_client.index(index="fhirstore.patient", body=patient_2)

        with open("test/fixtures/patient-example-with-extensions.json") as h:
            patient_3 = json.load(h)
            es_client.index(index="fhirstore.patient", body=patient_3)

    while es_client.count(index="fhirstore.patient")["count"] < 3:
        sleep(5 / 10000)
    return es_client


@pytest.fixture(scope="module")
def insert_medicationrequest(es_client):
    if not es_client.indices.exists("fhirstore.medicationrequest"):
        with open("test/fixtures/medicationrequest-example.json") as i:
            medicationrequest_1 = json.load(i)
            es_client.index(index="fhirstore.medicationrequest", body=medicationrequest_1)

    while es_client.count(index="fhirstore.medicationrequest")["count"] < 1:
        sleep(5 / 10000)
    return es_client


def test_search_output_type(store: FHIRStore, insert_patient):
    """Check that the output type is correct
    """
    result = store.comprehensive_search("Patient", ImmutableMultiDict([]))
    assert result.content["resource_type"] == "Bundle"


def test_search_medicationrequest(store: FHIRStore, insert_medicationrequest):
    """Check that medicationrequest was inserted properly
    """
    result = store.comprehensive_search("MedicationRequest", {})
    assert result.content["total"] == 1


def test_search_no_parameters(store: FHIRStore):
    """Checks that all elements of the resource are returned
    """
    result = store.comprehensive_search("Patient", ImmutableMultiDict([]))
    assert len(result.content["entry"]) == 3


def test_search_one_param_simple(store: FHIRStore):
    """Checks simple one parameter search
    """
    result = store.comprehensive_search(
        "Patient", ImmutableMultiDict([("identifier.value", "654321")])
    )
    assert len(result.content["entry"]) == 1
    assert all(
        element["resource"]["identifier"][0]["value"] == "654321"
        for element in result.content["entry"]
    )


def test_search_one_param_multiple(store: FHIRStore):
    """Checks that multiple elements of one parameter are queried
    """
    result = store.comprehensive_search(
        "Patient", ImmutableMultiDict([("name.family", "Chalmers,Levin")])
    )
    assert len(result.content["entry"]) == 2
    assert any(
        element["resource"]["name"][0]["family"] == ("Chalmers" or "Levin")
        for element in result.content["entry"]
    )
    assert all(
        element["resource"]["name"][0]["family"] != "Donald" for element in result.content["entry"]
    )


def test_search_one_param_modifier_str_contains(store: FHIRStore):
    """Checks that "contains" string modifier works
    """
    result = store.comprehensive_search(
        "Patient", ImmutableMultiDict([("managingOrganization.reference:contains", "Organization")])
    )
    assert len(result.content["entry"]) == 3
    assert any(
        element["resource"]["managingOrganization"]["reference"] == "Organization/1"
        for element in result.content["entry"]
    )
    assert any(
        element["resource"]["managingOrganization"]["reference"]
        != "Organization/2.16.840.1.113883.19.5"
        for element in result.content["entry"]
    )


def test_search_one_param_modifier_str_exact(store: FHIRStore):
    """Checks that "exact" string modifier works
    """
    result = store.comprehensive_search(
        "Patient", ImmutableMultiDict([("name.family:exact", "Donald")])
    )
    assert len(result.content["entry"]) == 1
    assert all(
        element["resource"]["name"][0]["family"] == "Donald" for element in result.content["entry"]
    )
    assert all(
        element["resource"]["name"][0]["family"] != ("Chalmers" or "Levin")
        for element in result.content["entry"]
    )


def test_search_one_param_modifier_str_not(store: FHIRStore):
    """Checks that "not" string modifier works
    """
    result = store.comprehensive_search(
        "Patient", ImmutableMultiDict([("name.family:not", "Donald")])
    )
    assert all(
        element["resource"]["name"][0]["family"] != "Donald" for element in result.content["entry"]
    )


def test_search_two_params_and(store: FHIRStore):
    """Checks two parameter "and" search
    """
    result = store.comprehensive_search(
        "Patient", ImmutableMultiDict([("name.family", "Chalmers"), ("identifier.value", "12345")])
    )
    assert all(
        element["resource"]["identifier"][0]["value"] == "12345"
        for element in result.content["entry"]
    )
    assert all(
        element["resource"]["identifier"][0]["value"] != "654321"
        for element in result.content["entry"]
    )
    assert all(
        element["resource"]["name"][0]["family"] == "Chalmers"
        for element in result.content["entry"]
    )
    assert all(
        element["resource"]["name"][0]["family"] != "Donald" or "Levin"
        for element in result.content["entry"]
    )
    assert len(result.content["entry"]) == 1


def test_search_one_params_and(store: FHIRStore):
    """Checks one parameter "and" search
    """
    result = store.comprehensive_search(
        "Patient", ImmutableMultiDict([("name.given", "Peter,James")])
    )
    assert len(result.content["entry"]) == 1
    assert any(
        element["resource"]["name"][0]["given"] == "Peter" or "James"
        for element in result.content["entry"]
    )
    assert all(
        element["resource"]["name"][0]["given"] != "Henry" for element in result.content["entry"]
    )
    assert all(
        element["resource"]["name"][0]["given"] != "Duck" for element in result.content["entry"]
    )


def test_search_and_or(store: FHIRStore):
    """Checks two parameter "and,or" search
    """
    result = store.comprehensive_search(
        "Patient",
        ImmutableMultiDict([("name.family", "Levin,Chalmers"), ("identifier.value", "12345")]),
    )
    assert len(result.content["entry"]) == 2
    assert all(
        element["resource"]["identifier"][0]["value"] != "654321"
        for element in result.content["entry"]
    )
    assert all(
        element["resource"]["identifier"][0]["value"] == "12345"
        for element in result.content["entry"]
    )
    assert all(
        element["resource"]["name"][0]["family"] == "Levin" or "Chalmers"
        for element in result.content["entry"]
    )
    assert all(
        element["resource"]["name"][0]["family"] != "Donald" for element in result.content["entry"]
    )


def test_search_identifier(store: FHIRStore):
    result = store.comprehensive_search(
        "Patient", ImmutableMultiDict([("managingOrganization:identifier", "98765")])
    )
    assert (
        result.content["entry"][0]["resource"]["managingOrganization"]["identifier"][0]["value"]
        == "98765"
    )


def test_count_all(store: FHIRStore):
    result = store.comprehensive_search("Patient", ImmutableMultiDict([("_summary", "count")]))
    assert result.content["total"] == 3
    assert result.content["tag"]["code"] == "SUBSETTED"


def test_count_medicationrequest(store: FHIRStore):
    result = store.comprehensive_search(
        "MedicationRequest", ImmutableMultiDict([("_summary", "count")])
    )
    assert result.content["total"] == 1


def test_count_some(store: FHIRStore):
    result = store.comprehensive_search(
        "Patient", ImmutableMultiDict([("_summary", "count"), ("identifier.value", "ne12345")])
    )
    assert result.content["total"] == 1
    assert result.content["tag"]["code"] == "SUBSETTED"


def test_search_element(store: FHIRStore):
    result = store.comprehensive_search(
        "Patient", ImmutableMultiDict([("_elements", "birthDate,gender")])
    )
    assert result.content["total"] == 3
    assert result.content["tag"]["code"] == "SUBSETTED"
    assert result.content["entry"] == [
        {"resource": {"gender": "male"}, "search": {"mode": "match"}},
        {"resource": {"gender": "male", "birthDate": "1932-09-24"}, "search": {"mode": "match"},},
        {"resource": {"gender": "male", "birthDate": "1974-12-25"}, "search": {"mode": "match"},},
    ]


def test_search_two_elements(store: FHIRStore):
    result = store.comprehensive_search("Patient", ImmutableMultiDict([("_elements", "birthDate")]))
    assert result.content["total"] == 3
    assert result.content["tag"]["code"] == "SUBSETTED"
    assert result.content["entry"] == [
        {"resource": {}, "search": {"mode": "match"}},
        {"resource": {"birthDate": "1932-09-24"}, "search": {"mode": "match"}},
        {"resource": {"birthDate": "1974-12-25"}, "search": {"mode": "match"}},
    ]


def test_search_summary_text(store: FHIRStore):
    result = store.comprehensive_search("Patient", ImmutableMultiDict([("_summary", "text")]))
    assert result.content["total"] == 3
    assert result.content["tag"]["code"] == "SUBSETTED"
    assert result.content["entry"] == [
        {
            "resource": {
                "meta": {
                    "tag": [
                        {
                            "system": "http://terminology.hl7.org/CodeSystem/v3-ActReason",
                            "code": "HTEST",
                            "display": "test health data",
                        }
                    ]
                },
                "id": "pat1",
                "text": {
                    "div": '<div xmlns="http://www.w3.org/1999/xhtml">\n      \n      <p>Patient Donald DUCK @ Acme Healthcare, Inc. MR = 654321</p>\n    \n    </div>',
                    "status": "generated",
                },
            },
            "search": {"mode": "match"},
        },
        {
            "resource": {
                "id": "xcda",
                "text": {
                    "div": '<div xmlns="http://www.w3.org/1999/xhtml">\n      \n      <p>Henry Levin the 7th</p>\n    \n    </div>',
                    "status": "generated",
                },
            },
            "search": {"mode": "match"},
        },
        {
            "resource": {
                "meta": {
                    "tag": [
                        {
                            "system": "http://terminology.hl7.org/CodeSystem/v3-ActReason",
                            "code": "HTEST",
                            "display": "test health data",
                        }
                    ]
                },
                "id": "example",
                "text": {
                    "div": '<div xmlns="http://www.w3.org/1999/xhtml">\n\t\t\t<table>\n\t\t\t\t<tbody>\n\t\t\t\t\t<tr>\n\t\t\t\t\t\t<td>Name</td>\n\t\t\t\t\t\t<td>Peter James \n              <b>Chalmers</b> (&quot;Jim&quot;)\n            </td>\n\t\t\t\t\t</tr>\n\t\t\t\t\t<tr>\n\t\t\t\t\t\t<td>Address</td>\n\t\t\t\t\t\t<td>534 Erewhon, Pleasantville, Vic, 3999</td>\n\t\t\t\t\t</tr>\n\t\t\t\t\t<tr>\n\t\t\t\t\t\t<td>Contacts</td>\n\t\t\t\t\t\t<td>Home: unknown. Work: (03) 5555 6473</td>\n\t\t\t\t\t</tr>\n\t\t\t\t\t<tr>\n\t\t\t\t\t\t<td>Id</td>\n\t\t\t\t\t\t<td>MRN: 12345 (Acme Healthcare)</td>\n\t\t\t\t\t</tr>\n\t\t\t\t</tbody>\n\t\t\t</table>\n\t\t</div>',
                    "status": "generated",
                },
            },
            "search": {"mode": "match"},
        },
    ]


def test_handle_pipe(store: FHIRStore):
    result = store.comprehensive_search(
        "MedicationRequest",
        ImmutableMultiDict([("contained.code.coding", "http://snomed.info/sct|324252006")]),
    )
    assert result.content["entry"][0]["resource"]["id"] == "medrx0302"
    assert (
        result.content["entry"][0]["resource"]["contained"][0]["code"]["coding"][0]["system"]
        == "http://snomed.info/sct"
    )
    assert (
        result.content["entry"][0]["resource"]["contained"][0]["code"]["coding"][0]["code"]
        == "324252006"
    )


def test_sort(store: FHIRStore):
    result = store.comprehensive_search("Patient", ImmutableMultiDict([("_sort", "birthDate")]))
    assert (
        result.content["entry"][0]["resource"]["birthDate"]
        <= result.content["entry"][1]["resource"]["birthDate"]
    )


def test_sort_desc(store: FHIRStore):
    result = store.comprehensive_search("Patient", ImmutableMultiDict([("_sort", "-birthDate")]))
    assert (
        result.content["entry"][0]["resource"]["birthDate"]
        >= result.content["entry"][1]["resource"]["birthDate"]
    )



import pytest
import json
from pytest import raises

from time import sleep
from fhirstore import FHIRStore, NotFoundError
from fhirstore.search import SearchArguments, parse_comma, pre_process_params
from collections.abc import Mapping
from werkzeug.datastructures import ImmutableMultiDict

# For now, this class assumes an already existing store exists
# (store.bootstrap was run)


@pytest.fixture(scope="module")
def insert_patient(es_client):
    if not es_client.indices.exists("fhirstore.patient"):
        with open("test/fixtures/patient-example.json") as f:
            patient_1 = json.load(f)
            es_client.index(index="fhirstore.patient", body=patient_1)

        with open("test/fixtures/patient-example-2.json") as g:
            patient_2 = json.load(g)
            es_client.index(index="fhirstore.patient", body=patient_2)

        with open("test/fixtures/patient-example-with-extensions.json") as h:
            patient_3 = json.load(h)
            es_client.index(index="fhirstore.patient", body=patient_3)

    while es_client.count(index="fhirstore.patient")["count"] < 3:
        sleep(5 / 10000)
    return es_client


@pytest.fixture(scope="module")
def insert_medicationrequest(es_client):
    if not es_client.indices.exists("fhirstore.medicationrequest"):
        with open("test/fixtures/medicationrequest-example.json") as i:
            medicationrequest_1 = json.load(i)
            es_client.index(index="fhirstore.medicationrequest", body=medicationrequest_1)

    while es_client.count(index="fhirstore.medicationrequest")["count"] < 1:
        sleep(5 / 10000)
    return es_client


def test_search_bad_resource_type(store: FHIRStore):
    """search() raises error if resource type is unknown"""

    with raises(NotFoundError, match='unsupported FHIR resource: "unknown"'):
        store.comprehensive_search(resource_type="unknown", args={})


def test_search_output_type(store: FHIRStore, insert_patient):
    """Check that the output type is correct
    """
    result = store.comprehensive_search("Patient", {})
    assert result.content["resource_type"] == "Bundle"


def test_search_medicationrequest(store: FHIRStore, insert_medicationrequest):
    """Check that medicationrequest was inserted properly
    """
    result = store.comprehensive_search("MedicationRequest", {})
    assert result.content["total"] == 1


def test_parse_comma_simple():
    key, value = parse_comma("gender", "male")
    assert key == "gender"
    assert value == ["male"]


def test_parse_comma_multiple():
    key, value = parse_comma("gender", "female,male")
    assert key == "multiple"
    assert value == {"gender": ["female", "male"]}


def test_no_params():
    resp = pre_process_params(ImmutableMultiDict([]))
    assert resp == {}


def test_one_param_no_comma():
    resp = pre_process_params(ImmutableMultiDict([("gender", "female")]))
    assert resp == {"gender": ["female"]}


def test_one_param_one_comma():
    resp = pre_process_params(ImmutableMultiDict([("gender", "female,male")]))
    assert resp == {"multiple": {"gender": ["female", "male"]}}


def test_one_param_two_entry_no_comma():

    resp = pre_process_params(ImmutableMultiDict([("name", "John"), ("name", "Lena")]))
    assert resp == {"name": ["John", "Lena"]}


def test_one_param_two_entries_one_comma():
    resp = pre_process_params(ImmutableMultiDict([("language", "FR"), ("language", "NL,EN")]))
    assert resp == {"language": ["FR"], "multiple": {"language": ["NL", "EN"]}}


def test_count_summary():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_summary", "count")]), "Patient")
    assert search_args.resource_type == "Patient"
    assert search_args.core_args == {}
    assert search_args.formatting_args["sort"] is None
    assert search_args.formatting_args["elements"] is None
    assert search_args.formatting_args["include"] is None
    assert search_args.formatting_args["summary"] is True
    assert search_args.formatting_args["is_summary_count"] is True
    assert search_args.meta_args["offset"] == 0
    assert search_args.meta_args["result_size"] == 100


def test_text_summary():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_summary", "text")]), "Patient")
    assert search_args.resource_type == "Patient"
    assert search_args.core_args == {}
    assert search_args.formatting_args["sort"] is None
    assert search_args.formatting_args["elements"] == ["text", "id", "meta"]
    assert search_args.formatting_args["include"] is None
    assert search_args.formatting_args["summary"] is True
    assert search_args.formatting_args["is_summary_count"] is False
    assert search_args.meta_args["offset"] == 0
    assert search_args.meta_args["result_size"] == 100


def test_element():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_elements", "birthDate")]), "Patient")
    assert search_args.resource_type == "Patient"
    assert search_args.core_args == {}
    assert search_args.formatting_args["sort"] is None
    assert search_args.formatting_args["elements"] == ["birthDate"]
    assert search_args.formatting_args["include"] is None
    assert search_args.formatting_args["summary"] is False
    assert search_args.formatting_args["is_summary_count"] is False
    assert search_args.meta_args["offset"] == 0
    assert search_args.meta_args["result_size"] == 100


def test_elements():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_elements", "birthDate,gender")]), "Patient")
    assert search_args.resource_type == "Patient"
    assert search_args.core_args == {}
    assert search_args.formatting_args["sort"] is None
    assert search_args.formatting_args["elements"] == ["birthDate", "gender"]
    assert search_args.formatting_args["include"] is None
    assert search_args.formatting_args["summary"] is False
    assert search_args.formatting_args["is_summary_count"] is False
    assert search_args.meta_args["offset"] == 0
    assert search_args.meta_args["result_size"] == 100


def test_result_size():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_count", "2")]), "Patient")
    assert search_args.resource_type == "Patient"
    assert search_args.core_args == {}
    assert search_args.formatting_args["sort"] is None
    assert search_args.formatting_args["elements"] is None
    assert search_args.formatting_args["include"] is None
    assert search_args.formatting_args["summary"] is False
    assert search_args.formatting_args["is_summary_count"] is False
    assert search_args.meta_args["offset"] == 0
    assert search_args.meta_args["result_size"] == 2


def test_result_size_elements():
    search_args = SearchArguments()
    search_args.parse(
        ImmutableMultiDict(
            [("_count", "2"), ("_summary", "False"), ("_elements", "birthDate,gender")]
        ),
        "Patient",
    )
    assert search_args.resource_type == "Patient"
    assert search_args.core_args == {}
    assert search_args.formatting_args["sort"] is None
    assert search_args.formatting_args["elements"] == ["birthDate", "gender"]
    assert search_args.formatting_args["include"] is None
    assert search_args.formatting_args["summary"] is True
    assert search_args.formatting_args["is_summary_count"] is False
    assert search_args.meta_args["offset"] == 0
    assert search_args.meta_args["result_size"] == 2


def test_mix_parameters():
    search_args = SearchArguments()
    search_args.parse(
        ImmutableMultiDict(
            [
                ("_count", "200"),
                ("_summary", "text"),
                ("_elements", "birthDate,gender"),
                ("language", "FR"),
                ("language", "EN,NL"),
            ]
        ),
        "Patient",
    )
    assert search_args.resource_type == "Patient"
    assert search_args.core_args == {
        "language": ["FR"],
        "multiple": {"language": ["EN", "NL"]},
    }
    assert search_args.formatting_args["sort"] is None
    assert search_args.formatting_args["elements"] == ["text", "id", "meta"]
    assert search_args.formatting_args["include"] is None
    assert search_args.formatting_args["summary"] is True
    assert search_args.formatting_args["is_summary_count"] is False
    assert search_args.meta_args["offset"] == 0
    assert search_args.meta_args["result_size"] == 200


def test_mix_params_count():
    search_args = SearchArguments()
    search_args.parse(
        ImmutableMultiDict([("_summary", "count"), ("language", "FR"), ("language", "EN,NL")]),
        "Patient",
    )
    assert search_args.resource_type == "Patient"
    assert search_args.core_args == {
        "language": ["FR"],
        "multiple": {"language": ["EN", "NL"]},
    }
    assert search_args.formatting_args["sort"] is None
    assert search_args.formatting_args["elements"] is None
    assert search_args.formatting_args["include"] is None
    assert search_args.formatting_args["summary"] is True
    assert search_args.formatting_args["is_summary_count"] is True
    assert search_args.meta_args["offset"] == 0
    assert search_args.meta_args["result_size"] == 100


def test_mix_params_sort():
    search_args = SearchArguments()
    search_args.parse(
        ImmutableMultiDict([("language", "FR"), ("language", "EN,NL"), ("_sort", "birthDate")]),
        "Patient",
    )
    assert search_args.core_args == {
        "language": ["FR"],
        "multiple": {"language": ["EN", "NL"]},
    }
    assert search_args.formatting_args["sort"] == ["birthDate"]
    assert search_args.formatting_args["elements"] is None
    assert search_args.formatting_args["include"] is None
    assert search_args.formatting_args["summary"] is False
    assert search_args.formatting_args["is_summary_count"] is False
    assert search_args.meta_args["offset"] == 0
    assert search_args.meta_args["result_size"] == 100


def test_mix_params_include():
    search_args = SearchArguments()
    search_args.parse(
        ImmutableMultiDict(
            [
                ("language", "FR"),
                ("language", "EN,NL"),
                ("_include", "Patient:managingOrganization"),
            ]
        ),
        "Patient",
    )
    assert search_args.core_args == {
        "language": ["FR"],
        "multiple": {"language": ["EN", "NL"]},
    }
    assert search_args.formatting_args["sort"] is None
    assert search_args.formatting_args["elements"] is None
    assert search_args.formatting_args["include"] == ["managingOrganization"]
    assert search_args.formatting_args["summary"] is False
    assert search_args.formatting_args["is_summary_count"] is False
    assert search_args.meta_args["offset"] == 0
    assert search_args.meta_args["result_size"] == 100


def test_mix_params_include_multiple():
    search_args = SearchArguments()
    search_args.parse(
        ImmutableMultiDict([("_include", "MedicationRequest:subject,MedicationRequest:requester")]),
        "Patient",
    )
    assert search_args.core_args == {}
    assert search_args.formatting_args["sort"] is None
    assert search_args.formatting_args["elements"] is None
    assert search_args.formatting_args["include"] == ["subject", "requester"]
    assert search_args.formatting_args["summary"] is False
    assert search_args.formatting_args["is_summary_count"] is False
    assert search_args.meta_args["offset"] == 0
    assert search_args.meta_args["result_size"] == 100


def test_sort_param():
    search_args = SearchArguments()
    search_args.parse({"_sort": ["birthDate"]}, "Patient")
    assert search_args.formatting_args["sort"] == ["birthDate"]


def test_no_sort_param():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("language", "FR")]), "Patient")
    assert search_args.formatting_args["sort"] is None


def test_sort_param_desc():
    search_args = SearchArguments()
    search_args.parse({"_sort": ["-birthDate"]}, "Patient")
    assert search_args.formatting_args["sort"] == [{"birthDate": {"order": "desc"}}]


def test_sort_score():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_sort", "_score")]), "Patient")
    assert search_args.formatting_args["sort"] == ["_score"]


def test_sort_score_asc():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_sort", "-_score")]), "Patient")
    assert search_args.formatting_args["sort"] == [{"_score": {"order": "asc"}}]


def test_sort_params():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_sort", "-birthDate,active")]), "Patient")
    assert search_args.formatting_args["sort"] == [{"birthDate": {"order": "desc"}}, "active"]


def test_sort_params_score():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_sort", "-birthDate,_score")]), "Patient")
    assert search_args.formatting_args["sort"] == [{"birthDate": {"order": "desc"}}, "_score"]


def test_sort_params_desc_and_score_asc():
    search_args = SearchArguments()
    search_args.parse(ImmutableMultiDict([("_sort", "-birthDate,-_score")]), "Patient")
    assert search_args.formatting_args["sort"] == [
        {"birthDate": {"order": "desc"}},
        {"_score": {"order": "asc"}},
    ]



import json
from pytest import fixture, raises
from unittest.mock import patch

from bson.objectid import ObjectId
from pymongo import MongoClient
from pymongo.errors import DuplicateKeyError
from jsonschema.exceptions import ValidationError
from collections.abc import Mapping

from fhirstore import FHIRStore, BadRequestError, NotFoundError, ARKHN_CODE_SYSTEMS


@fixture(autouse=True)
def reset_store(store):
    store.reset()
    store.bootstrap(depth=2, resource="Patient")
    store.bootstrap(depth=2, resource="MedicationRequest")


# For now, this class assumes an already existing store exists
# (store.bootstrap was run)


class TestFHIRStore:
    "FHIRStore"

    ###
    # FHIRStore.create()
    ###
    def test_create_missing_resource_type(self, store: FHIRStore):
        """create() raises if resource type is not specified"""

        with raises(BadRequestError, match="resourceType is missing in resource"):
            store.create({})

    def test_create_bad_resource_type(self, store: FHIRStore):
        """create() raises if resource type is unknown"""

        with raises(NotFoundError, match='unsupported FHIR resource: "unknown"'):
            store.create({"resourceType": "unknown"})

    def test_create_bad_resource_schema(self, store: FHIRStore):
        """create() raises if json schema validation failed in mongo"""

        with raises(ValidationError):
            store.create({"resourceType": "Patient", "id": 42})

    def test_create_resource(self, store: FHIRStore, mongo_client: MongoClient, test_patient):
        """create() correctly inserts a document in the database"""
        result = store.create(test_patient)
        assert isinstance(result["_id"], ObjectId), "result _id must be an objectId"
        inserted = mongo_client["Patient"].find_one({"_id": result["_id"]})
        assert inserted == test_patient

    def test_create_resource_with_extension(self, store: FHIRStore, mongo_client: MongoClient):
        """resources using extensions are not
        handled yet, an error should be raised"""
        with open("test/fixtures/patient-example-with-extensions.json") as f:
            patient = json.load(f)
            result = store.create(patient)
            assert isinstance(result["_id"], ObjectId), "result _id must be an objectId"
            inserted = mongo_client["Patient"].find_one({"_id": result["_id"]})
            assert inserted == patient
            mongo_client["Patient"].delete_one({"_id": patient["_id"]})

    def test_unique_indices(self, store: FHIRStore, mongo_client: MongoClient):
        """create() raises if index is already present"""
        store.create(
            {
                "identifier": [
                    {"value": "val", "system": "sys"},
                    {"value": "just_value"},
                    {
                        "value": "value",
                        "system": "system",
                        "type": {"coding": [{"system": "type_system", "code": "type_code"}]},
                    },
                ],
                "resourceType": "Patient",
                "id": "pat1",
            }
        )

        # index on id
        with raises(DuplicateKeyError, match='dup key: { id: "pat1" }'):
            store.create({"resourceType": "Patient", "id": "pat1"})

        # index on (identifier.value, identifier.system)
        with raises(
            DuplicateKeyError,
            match='dup key: { identifier.system: "sys", identifier.value: "val", \
identifier.type.coding.0.system: null, identifier.type.coding.0.code: null }',
        ):
            store.create(
                {
                    "identifier": [{"value": "val", "system": "sys"}],
                    "resourceType": "Patient",
                    "id": "pat2",
                }
            )
        with raises(
            DuplicateKeyError,
            match='dup key: { identifier.system: null, identifier.value: "just_value", \
identifier.type.coding.0.system: null, identifier.type.coding.0.code: null }',
        ):
            store.create(
                {"identifier": [{"value": "just_value"}], "resourceType": "Patient", "id": "pat2",}
            )
        with raises(
            DuplicateKeyError,
            match='dup key: { identifier.system: "system", identifier.value: "value", \
identifier.type.coding.0.system: "type_system", identifier.type.coding.0.code: "type_code" }',
        ):
            store.create(
                {
                    "identifier": [
                        {"value": "new_val"},
                        {
                            "value": "value",
                            "system": "system",
                            "type": {"coding": [{"system": "type_system", "code": "type_code"}]},
                        },
                    ],
                    "resourceType": "Patient",
                    "id": "pat2",
                }
            )

    ###
    # FHIRStore.read()
    ###

    def test_read_bad_resource_type(self, store: FHIRStore):
        """read() raises if resource type is unknown"""

        with raises(NotFoundError, match='unsupported FHIR resource: "unknown"'):
            store.read("unknown", "864321")

    def test_read_resource_not_found(self, store: FHIRStore, test_patient):
        """read() returns None when no matching document was found"""

        with raises(NotFoundError):
            store.read("Patient", test_patient["id"])

    def test_read_resource(self, store: FHIRStore, test_patient):
        """read() finds a document in the database"""
        store.create(test_patient)
        result = store.read("Patient", test_patient["id"])
        assert result == test_patient

    ###
    # FHIRStore.update()
    ###
    def test_update_bad_resource_type(self, store: FHIRStore):
        """update() raises if resource type is unknown"""

        with raises(NotFoundError, match='unsupported FHIR resource: "unknown"'):
            store.update("unknown", "864321", {"gender": "other"})

    def test_update_resource_not_found(self, store: FHIRStore, test_patient):
        """update() returns None when no matching document was found"""

        with raises(NotFoundError):
            store.update("Patient", test_patient["id"], {"name": "Patator"})

    def test_update_bad_resource_schema(self, store: FHIRStore, test_patient):
        """update() raises if json schema validation failed in mongo"""

        store.create(test_patient)
        with raises(ValidationError):
            store.update("Patient", test_patient["id"], {**test_patient, "gender": "elephant"})

    def test_update_resource(self, store: FHIRStore, test_patient):
        """update() finds a document in the database"""
        store.create(test_patient)
        store.update("Patient", test_patient["id"], {**test_patient, "gender": "other"})
        assert store.read("Patient", test_patient["id"]) == {**test_patient, "gender": "other"}

    ###
    # FHIRStore.patch()
    ###
    def test_patch_bad_resource_type(self, store: FHIRStore):
        """patch() raises if resource type is unknown"""

        with raises(NotFoundError, match='unsupported FHIR resource: "unknown"'):
            store.patch("unknown", "864321", {"gender": "other"})

    def test_patch_resource_not_found(self, store: FHIRStore, test_patient):
        """patch() returns None when no matching document was found"""

        with raises(NotFoundError):
            store.patch("Patient", test_patient["id"], {"name": "Patator"})

    def test_patch_bad_resource_schema(self, store: FHIRStore, test_patient):
        """patch() raises if json schema validation failed in mongo"""

        store.create(test_patient)
        with raises(ValidationError):
            store.patch("Patient", test_patient["id"], {"gender": "elephant"})

    def test_patch_resource(self, store: FHIRStore, test_patient):
        """patch() finds a document in the database"""
        store.create(test_patient)
        store.patch("Patient", test_patient["id"], {"gender": "other"})
        assert store.read("Patient", test_patient["id"]) == {**test_patient, "gender": "other"}

    ###
    # FHIRStore.delete()
    ###
    def test_delete_bad_resource_type(self, store: FHIRStore):
        """delete() raises if resource type is unknown"""

        with raises(NotFoundError, match='unsupported FHIR resource: "unknown"'):
            store.delete("unknown", "864321")

    def test_delete_resource_not_found(self, store: FHIRStore, test_patient):
        """delete() returns None when no matching document was found"""

        with raises(NotFoundError):
            store.delete("Patient", test_patient["id"])

    def test_delete_missing_param(self, store: FHIRStore, test_patient):
        """delete() returns None when no matching document was found"""

        with raises(
            BadRequestError,
            match="one of: 'instance_id', 'resource_id' or 'source_id' are required",
        ):
            store.delete("Patient")

    def test_delete_instance(self, store: FHIRStore, test_patient):
        """delete() finds a document in the database"""
        store.create(test_patient)
        result = store.delete("Patient", test_patient["id"])
        assert result == 1

    def test_delete_by_resource_id(self, store: FHIRStore, test_patient):
        """delete() finds a document in the database"""
        store.create(test_patient)

        resource_id = "pyrogResourceId"
        metadata = {
            "tag": [
                {"system": ARKHN_CODE_SYSTEMS.resource, "code": resource_id},
                {"code": "some-other-tag"},
            ]
        }
        store.create({"resourceType": "Patient", "id": "pat2", "meta": metadata})
        store.create({"resourceType": "Patient", "id": "pat3", "meta": metadata})

        result = store.delete("Patient", resource_id=resource_id)
        assert result == 2

    def test_delete_by_source_id(self, store: FHIRStore, test_patient):
        """delete() finds a document in the database"""
        store.create(test_patient)

        source_id = "pyrogSourceId"
        metadata = {
            "tag": [
                {"system": ARKHN_CODE_SYSTEMS.source, "code": source_id},
                {"code": "some-other-tag"},
            ]
        }
        store.create({"resourceType": "Patient", "id": "pat2", "meta": metadata})
        store.create({"resourceType": "Patient", "id": "pat3", "meta": metadata})

        result = store.delete("Patient", source_id=source_id)
        assert result == 2

    ###
    # FHIRStore.upload_bundle()
    ###
    @patch("fhirstore.FHIRStore.create")
    def test_upload_bundle(
        self, mock_create, store: FHIRStore, mongo_client: MongoClient, test_bundle
    ):
        """create() correctly inserts a document in the database"""
        store.upload_bundle(test_bundle)
        assert mock_create.call_count == 3



import sys
import logging
import elasticsearch

from pymongo import MongoClient, ASCENDING
from pymongo.errors import WriteError, OperationFailure, DuplicateKeyError
from tqdm import tqdm
from jsonschema import validate

from fhirstore import ARKHN_CODE_SYSTEMS
from fhirstore.schema import SchemaParser
from fhirstore.search_engine import ElasticSearchEngine, create_search_engine

from fhirpath.search import SearchContext, Search


class NotFoundError(Exception):
    """
    NotFoundError is returned when a resource was not found in the database.
    """

    pass


class BadRequestError(Exception):
    """
    BadRequestError is returned when the client request could not be processed.
    """

    pass


class FHIRStore:
    def __init__(
        self,
        mongo_client: MongoClient,
        es_client: elasticsearch.Elasticsearch,
        db_name: str,
        resources: dict = {},
    ):
        self.es = es_client
        self.db = mongo_client[db_name]
        self.parser = SchemaParser()
        self.resources = resources

        if self.es and len(self.es.transport.hosts) > 0:
            self.search_engine: ElasticSearchEngine = create_search_engine(self.es)
        else:
            logging.warning("No elasticsearch client provided, search features are disabled")

    def reset(self):
        """
        Drops all collections currently in the database.
        """
        for collection in self.db.list_collection_names():
            self.db.drop_collection(collection)
        self.es.indices.delete("_all")
        self.resources = {}

    def bootstrap(self, depth=3, resource=None, show_progress=True):
        """
        Parses the FHIR json-schema and create the collections according to it.
        """
        # Bootstrap elastic indices
        self.search_engine.create_es_index()

        # Bootstrap mongoDB collections
        resources = self.parser.parse(depth=depth, resource=resource)
        if show_progress:
            tqdm.write("\n", end="")
            resources = tqdm(
                resources,
                file=sys.stdout,
                desc="Bootstrapping collections...",
                total=len(self.search_engine.mappings),
            )
        for resource_name, schema in resources:
            self.db.create_collection(resource_name, **{"validator": {"$jsonSchema": schema}})
            # Add unique constraint on id
            self.db[resource_name].create_index("id", unique=True)
            # Add unique constraint on (identifier.system, identifier.value)
            self.db[resource_name].create_index(
                [
                    ("identifier.system", ASCENDING),
                    ("identifier.value", ASCENDING),
                    ("identifier.type.coding.0.system", ASCENDING),
                    ("identifier.type.coding.0.code", ASCENDING),
                ],
                unique=True,
                partialFilterExpression={"identifier": {"$exists": True}},
            )
            self.resources[resource_name] = schema

    def resume(self, show_progress=True):
        """
        Loads the existing resources schema from the database.
        """
        collections = self.db.list_collection_names()

        if show_progress:
            tqdm.write("\n", end="")
            collections = tqdm(
                collections, file=sys.stdout, desc="Loading collections from database...",
            )

        for collection in collections:
            self.resources[collection] = {}
        # for collection in collections:
        #     json_schema = self.db.get_collection(collection).options()["validator"]["$jsonSchema"]
        #     self.resources[collection] = json_schema

    def validate_resource_type(self, resource_type):
        if resource_type is None:
            raise BadRequestError("resourceType is missing in resource")

        elif resource_type not in self.resources:
            raise NotFoundError(f'unsupported FHIR resource: "{resource_type}"')

    def create(self, resource, bypass_document_validation=False):
        """
        Creates a resource. The structure of the resource will be checked
        against its json-schema FHIR definition.

        Args:
            - resource_type: type of the resource (eg: 'Patient')
            - id: The expected id is the resource 'id', not the
                  internal database identifier ('_id').

        Returns: The updated resource.
        """
        resource_type = resource.get("resourceType")
        self.validate_resource_type(resource_type)

        try:
            res = self.db[resource_type].insert_one(
                resource, bypass_document_validation=bypass_document_validation
            )
            return {**resource, "_id": res.inserted_id}
        except DuplicateKeyError as e:
            raise e
        except WriteError:
            self.validate(resource)

    def read(self, resource_type, instance_id):
        """
        Finds a resource given its type and id.

        Args:
            - resource_type: type of the resource (eg: 'Patient')
            - id: The expected id is the resource 'id', not the
                  internal database identifier ('_id').

        Returns: The found resource.
        """
        self.validate_resource_type(resource_type)

        res = self.db[resource_type].find_one({"id": instance_id})
        if res is None:
            raise NotFoundError
        return res

    def update(self, resource_type, instance_id, resource, bypass_document_validation=False):
        """
        Update a resource given its type, id and a resource. It applies
        a "replace" operation, therefore the resource will be overriden.
        The structure of the updated resource will  be checked against
        its json-schema FHIR definition.

        Args:
            - resource_type: type of the resource (eg: 'Patient')
            - id: The expected id is the resource 'id', not the
                  internal database identifier ('_id').
            - resource: The updated resource.

        Returns: The updated resource.
        """
        self.validate_resource_type(resource_type)

        try:
            update_result = self.db[resource_type].replace_one(
                {"id": instance_id},
                resource,
                bypass_document_validation=bypass_document_validation,
            )
            if update_result.matched_count == 0:
                raise NotFoundError
            return update_result
        except OperationFailure:
            self.validate(resource)

    def patch(self, resource_type, instance_id, patch, bypass_document_validation=False):
        """
        Update a resource given its type, id and a patch. It applies
        a "patch" operation rather than a "replace", only the fields
        specified in the third argument will be updated. The structure
        of the updated resource will  be checked against its json-schema
        FHIR definition.

        Args:
            - resource_type: type of the resource (eg: 'Patient')
            - id: The expected id is the resource 'id', not the
                  internal database identifier ('_id').
            - patch: The patch to be applied on the resource.

        Returns: The updated resource.
        """
        self.validate_resource_type(resource_type)

        try:
            update_result = self.db[resource_type].update_one(
                {"id": instance_id},
                {"$set": patch},
                bypass_document_validation=bypass_document_validation,
            )
            if update_result.matched_count == 0:
                raise NotFoundError
            return update_result
        except OperationFailure:
            resource = self.read(resource_type, instance_id)
            self.validate({**resource, **patch})

    def delete(self, resource_type, instance_id=None, resource_id=None, source_id=None):
        """
        Deletes a resource given its type and id.

        Args:
            - resource_type: type of the resource (eg: 'Patient')
            - id: The expected id is the resource 'id', not the
                  internal database identifier ('_id').

        Returns: The id of the deleted resource.
        """
        self.validate_resource_type(resource_type)

        if instance_id:
            res = self.db[resource_type].delete_one({"id": instance_id})
        elif resource_id:
            res = self.db[resource_type].delete_many(
                {
                    "meta.tag": {
                        "$elemMatch": {
                            "code": {"$eq": resource_id},
                            "system": {"$eq": ARKHN_CODE_SYSTEMS.resource},
                        }
                    }
                }
            )
        elif source_id:
            res = self.db[resource_type].delete_many(
                {
                    "meta.tag": {
                        "$elemMatch": {
                            "code": {"$eq": source_id},
                            "system": {"$eq": ARKHN_CODE_SYSTEMS.source},
                        }
                    }
                }
            )
        else:
            raise BadRequestError(
                "one of: 'instance_id', 'resource_id' or 'source_id' are required"
            )

        if res.deleted_count == 0:
            raise NotFoundError

        return res.deleted_count

    def validate(self, resource):
        """
        Validates the given resource against its own schema.
        This is much more efficient than running th validation against the
        whole FHIR json schema.
        This function is useful because MongoDB does not provide any feedback
        about the schema validation error other than "Schema validation failed"

        Args:
            resource: The object to be validated against the schema.
                      It is expected to have the "resourceType" property.
        """
        schema = self.resources.get(resource["resourceType"])
        if schema is None:
            raise Exception(f"missing schema for resource {resource}")
        validate(instance=resource, schema=schema)

    def search(self, resource_type, query_string=None, params=None):
        """
        Searchs for params inside a resource.
        Returns a bundle of items, as required by FHIR standards.

        Args:
            - resource_type: FHIR resource (eg: 'Patient')
            - params: search parameters as returned by the API. For a simple
            search, the parameters should be of the type {"key": "value"}
            eg: {"gender":"female"}, with possible modifiers {"address.city:exact":"Paris"}.
            If a search is made one field with multiple arguments (eg: language is French 
            OR English), params should be a payload of type {"multiple": {"language": 
            ["French", "English"]}}. 
            If a search has more than one field queried, params should be a payload of 
            the form: {"address.city": ["Paris"], "multiple":
            {"language": ["French", "English"]}}.
        Returns: A bundle with the results of the search, as required by FHIR
        search standard.
        """
        # self.validate_resource_type(resource_type)
        search_context = SearchContext(self.search_engine, resource_type)
        fhir_search = Search(search_context, query_string=query_string, params=params)

        return fhir_search()

    # def comprehensive_search(self, resource_type: str, args: ImmutableMultiDict):
    #     """ To deal with keywords : _include, _revinclude, _has
    #     """
    #     search_args = SearchArguments()
    #     search_args.parse(args, resource_type)

    #     # handle _has
    #     rev_chain = search_args.reverse_chain
    #     if rev_chain and rev_chain.is_queried:
    #         inner_ids = []
    #         # If there is a double _has chain
    #         if len(rev_chain.has_args) == 2:
    #             outer_ids = []
    #             rev_args_outer = SearchArguments()

    #             # parse the outer chain and search the store
    #             rev_args_outer.parse(rev_chain.has_args[1], rev_chain.resources_type[1])
    #             chained_bundle_outer = self.search(rev_args_outer)

    #             for item in chained_bundle_outer.content["entry"]:
    #                 outer_ids.append(
    #                     item["resource"][rev_chain.references[1]]["reference"].split(
    #                         sep="/", maxsplit=1
    #                     )[1]
    #                 )

    #             # fill the inner chain with the ids from the previous search
    #             rev_chain.has_args[0][rev_chain.fields[0]] = outer_ids

    #         # If there is a single _has chain, fill it with the value
    #         elif len(rev_chain.has_args) == 1:
    #             rev_chain.has_args[0][rev_chain.fields[0]] = rev_chain.value

    #         rev_args_inner = SearchArguments()
    #         rev_args_inner.parse(rev_chain.has_args[0], rev_chain.resources_type[0])
    #         chained_bundle_inner = self.search(rev_args_inner)

    #         for item in chained_bundle_inner.content["entry"]:
    #             inner_ids.append(
    #                 item["resource"][rev_chain.references[0]]["reference"].split(
    #                     sep="/", maxsplit=1
    #                 )[1]
    #             )
    #         search_args.core_args["id"] = inner_ids

    #     bundle = self.search(search_args)

    #     # handle _include
    #     if (
    #         search_args.formatting_args["include"]
    #         and not search_args.formatting_args["is_summary_count"]
    #     ):
    #         included_hits = {}
    #         for item in bundle.content["entry"]:
    #             # only go over items that are not the result of an inclusion
    #             if item["search"]["mode"] == "include":
    #                 continue
    #             # For each attribute to include
    #             for attribute in search_args.formatting_args["include"]:
    #                 # split the reference attribute "Practioner/123" into a
    #                 # resource "Practioner" and an id "123"
    #                 try:
    #                     included_resource, included_id = item["resource"][attribute][
    #                         "reference"
    #                     ].split(sep="/", maxsplit=1)
    #                     included_hits = self.es.search(
    #                         body={
    #                             "query": {
    #                                 "simple_query_string": {"query": included_id, "fields": ["id"]}
    #                             }
    #                         },
    #                         index=f"fhirstore.{included_resource.lower()}",
    #                     )
    #                 except KeyError as e:
    #                     logging.warning(f"Attribute: {e} is empty")
    #                 except elasticsearch.exceptions.NotFoundError as e:
    #                     logging.warning(
    #                         f"{e.info['error']['index']} is not indexed in the database yet."
    #                     )

    #         bundle.append(included_hits, search_args.formatting_args)

    #     return bundle

    def upload_bundle(self, bundle):
        """
        Upload a bundle of resource instances to the store.

        Args:
            - bundle: the fhir bundle containing the resources.
        """
        if "resourceType" not in bundle or bundle["resourceType"] != "Bundle":
            raise Exception("input must be a FHIR Bundle resource")

        for entry in bundle["entry"]:
            if "resource" not in entry:
                raise Exception("Bundle entry is missing a resource.")

            try:
                self.create(entry["resource"])
            except DuplicateKeyError as e:
                logging.warning(f"Document already existed: {e}")



import os
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


requirements = read("requirements.txt").split("\n")

setuptools.setup(
    name="fhirstore",
    version="0.5.0-dev1",
    author="Arkhn",
    author_email="contact@arkhn.org",
    description="Manipulating FHIR data leveraging MongoDB as storage layer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/arkhn/pyfhirstore/",
    packages=setuptools.find_packages(),
    package_data={"fhirstore": ["schema/fhir.schema.json"]},
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import pytest
import json
from pytest import raises

from fhirstore.search import build_simple_query, build_core_query, build_element_query
from collections.abc import Mapping
from fhirstore import FHIRStore, NotFoundError


######
## TEST BUILD SIMPLE QUERY
##
######


def test_basic_int_query():
    result = build_simple_query({"value": [3]})
    assert result == {"match": {"value": 3}}


def test_basic_float_query():
    result = build_simple_query({"value": [3.5]})
    assert result == {"match": {"value": 3.5}}


def test_simple_query_contains():
    """Validates that the ES query is correct with modifier contains
    """
    result = build_simple_query({"name.family:contains": ["Dona"]})
    assert result == {"query_string": {"query": "*Dona*", "fields": ["name.family"]}}


def test_simple_query_exact():
    """Validates that the ES query is correct with modifier contains
    """
    result = build_core_query({"name.family:exact": ["Donald"]})
    assert result == {
        "simple_query_string": {"query": '"Donald"', "fields": ["name.family"], "flags": "PHRASE"}
    }


def test_simple_query_not():
    """Validates that the ES query is correct with modifier contains
    """
    result = build_core_query({"name.family:not": ["Donald"]})
    assert result == {"simple_query_string": {"query": "-Donald", "fields": ["name.family"]}}


def test_simple_query_identifier():
    """Validates that the ES query is correct with modifier contains
    """
    result = build_simple_query(
        {"managingOrganization:identifier": ["urn:oid:0.7.6.5.4.3.2|98765"]}
    )
    assert result == {
        "simple_query_string": {
            "query": '"urn:oid:0.7.6.5.4.3.2|98765"',
            "fields": ["managingOrganization.identifier.value"],
            "flags": "PHRASE",
        }
    }


def test_simple_query_gt():
    """Validates that the ES query is correct with modifier 'gt'
    """
    result = build_simple_query({"birthDate": ["gt1974-12-25"]})
    assert result == {"range": {"birthDate": {"gt": "1974-12-25"}}}


def test_simple_query_ge():
    """Validates that the ES query is correct with modifier 'ge'
    """
    result = build_simple_query({"birthDate": ["ge1974-12-25"]})
    assert result == {"range": {"birthDate": {"gte": "1974-12-25"}}}


def test_simple_query_le():
    """Validates that the ES query is correct with modifier 'le'
    """
    result = build_simple_query({"birthDate": ["le1974-12-25"]})
    assert result == {"range": {"birthDate": {"lte": "1974-12-25"}}}


def test_simple_query_lt():
    """Validates that the ES query is correct with modifier 'lt'
    """
    result = build_simple_query({"birthDate": ["lt1974-12-25"]})
    assert result == {"range": {"birthDate": {"lt": "1974-12-25"}}}


def test_simple_query_eq():
    """Validates that the ES query is correct with modifier 'eq'
    """
    result = build_simple_query({"birthDate": ["eq1974-12-25"]})
    assert result == {"match": {"birthDate": "1974-12-25"}}


def test_simple_query_ne():
    """Validates that the ES query is correct with modifier 'gt'
    """
    result = build_core_query({"birthDate": ["ne1974-12-25"]})
    assert result == {"simple_query_string": {"query": "-1974-12-25", "fields": ["birthDate"]}}


def test_composite_query_pipe():
    """Validates that the ES query with token | is correct
    """
    result = build_simple_query({"contained.code.coding": ["http://snomed.info/sct|324252006"]})
    assert result == {
        "bool": {
            "must": [
                {
                    "simple_query_string": {
                        "query": '"http://snomed.info/sct"',
                        "fields": ["contained.code.coding.system"],
                        "flags": "PHRASE",
                    }
                },
                {
                    "simple_query_string": {
                        "query": '"324252006"',
                        "fields": ["contained.code.coding.code", "contained.code.coding.value"],
                        "flags": "PHRASE",
                    }
                },
            ]
        }
    }


def test_simple_query_string():
    """Validates that the ES query is correct"""
    result = build_simple_query({"birthDate": ["1974-12-25"]})
    assert result == {
        "simple_query_string": {"fields": ["birthDate"], "query": '"1974-12-25"', "flags": "PHRASE"}
    }


def test_simple_query_output():
    """Validates that the output of build_simple_query
    is a dictionary
    """
    result = build_simple_query({"gender": ["female"]})
    assert isinstance(result, Mapping)


def test_element_query_output():
    """Validates that the output of build_simple_query
    is a dictionary
    """
    element_query = build_element_query("gender", "female")
    assert isinstance(element_query, Mapping)


##########
##
# Test build_core_query
##
##########


def test_core_query_builder_validate():
    with raises(AssertionError, match="parameters must be a dictionary"):
        build_core_query(["name"])


def test_core_query_builder():
    corequerybuilder = build_core_query({})
    assert corequerybuilder == {"match_all": {}}


def test_core_query_builder_int():
    corequerybuilder = build_core_query({"value": [3]})
    assert corequerybuilder == {"match": {"value": 3}}


def test_core_query_builder_float():
    corequerybuilder = build_core_query({"value": [3.5]})
    assert corequerybuilder == {"match": {"value": 3.5}}


def test_core_query_correct_string():
    """Validates that the ES query is correct"""
    corequerybuilder = build_core_query({"birthDate": ["1974-12-25"]})
    assert corequerybuilder == {
        "simple_query_string": {"fields": ["birthDate"], "query": '"1974-12-25"', "flags": "PHRASE"}
    }


def test_core_query_multiple():
    corequerybuilder = build_core_query({"multiple": {"language": ["NL", "EN"]}})
    assert corequerybuilder == {
        "bool": {
            "should": [
                {
                    "simple_query_string": {
                        "query": '"NL"',
                        "fields": ["language"],
                        "flags": "PHRASE",
                    }
                },
                {
                    "simple_query_string": {
                        "query": '"EN"',
                        "fields": ["language"],
                        "flags": "PHRASE",
                    }
                },
            ]
        }
    }


def test_core_query_composite():
    corequerybuilder = build_core_query(
        {"language": ["FR"], "multiple": {"language": ["NL", "EN"]}}
    )
    assert corequerybuilder == {
        "bool": {
            "must": [
                {
                    "simple_query_string": {
                        "query": '"FR"',
                        "fields": ["language"],
                        "flags": "PHRASE",
                    }
                },
                {
                    "bool": {
                        "should": [
                            {
                                "simple_query_string": {
                                    "query": '"NL"',
                                    "fields": ["language"],
                                    "flags": "PHRASE",
                                }
                            },
                            {
                                "simple_query_string": {
                                    "query": '"EN"',
                                    "fields": ["language"],
                                    "flags": "PHRASE",
                                }
                            },
                        ]
                    }
                },
            ]
        }
    }


def test_simple_query_correct_or():
    """Validates that the ES query for "OR" is correct
    """
    result = build_simple_query({"multiple": {"name.family": ["Donald", "Chalmers"]}})
    assert result == {
        "bool": {
            "should": [
                {
                    "simple_query_string": {
                        "query": '"Donald"',
                        "fields": ["name.family"],
                        "flags": "PHRASE",
                    }
                },
                {
                    "simple_query_string": {
                        "query": '"Chalmers"',
                        "fields": ["name.family"],
                        "flags": "PHRASE",
                    }
                },
            ]
        }
    }


def test_simple_query_correct_and():
    """Validates that the ES query for "AND" is correct
    """
    result = build_simple_query({"name.family": ["Donald", "Chalmers"]})
    assert result == {
        "bool": {
            "must": [
                {
                    "simple_query_string": {
                        "fields": ["name.family"],
                        "query": '"Donald"',
                        "flags": "PHRASE",
                    }
                },
                {
                    "simple_query_string": {
                        "fields": ["name.family"],
                        "query": '"Chalmers"',
                        "flags": "PHRASE",
                    }
                },
            ]
        }
    }


def test_composite_query_modifiers():
    """Validates that the ES query for "AND" with modifiers is correct
    """
    result = build_simple_query({"birthDate": ["lt20200101", "gt20100101"]})
    assert result == {
        "bool": {
            "must": [
                {"range": {"birthDate": {"lt": "20200101"}}},
                {"range": {"birthDate": {"gt": "20100101"}}},
            ]
        }
    }


def test_core_query_output():
    """Validates that the output of build_element_query
    is a dictionary
    """
    result = build_simple_query({"gender": ["female"]})
    assert isinstance(result, Mapping)



from fhirstore.search import SearchArguments


class Bundle:
    def __init__(self):
        self.content = {"resource_type": "Bundle", "total": 0, "entry": []}

    def fill(self, hits, formatting_args):
        if formatting_args["is_summary_count"] == True:
            self.content["tag"] = {"code": "SUBSETTED"}
            self.content.pop("entry")
            if len(hits):
                self.content["total"] += hits["count"]

        elif formatting_args["is_summary_count"] == False and len(hits):
            for h in hits["hits"]["hits"]:
                self.content["entry"].append(
                    {"resource": h["_source"], "search": {"mode": "match"}}
                )
            self.content["total"] += hits["hits"]["total"]["value"]
        if formatting_args["elements"] or formatting_args["summary"]:
            self.content["tag"] = {"code": "SUBSETTED"}

    def complete(self, new_bundle, formatting_args):
        self.content["total"] += new_bundle.content["total"]
        if not formatting_args["is_summary_count"]:
            self.content["entry"].extend(new_bundle.content["entry"])

    def append(self, included_hits, formatting_args):
        if formatting_args["include"] and "hits" in included_hits:
            for h in included_hits["hits"]["hits"]:
                self.content["entry"].append(
                    {"resource": h["_source"], "search": {"mode": "include"}}
                )

    def fill_error(self, severity="error", code="invalid", details=None, diagnostic=None):
        self.content["resource_type"] = "OperationOutcome"
        self.content["issue"] = {"severity": severity, "code": code}
        self.content.pop("total", None)
        self.content.pop("entry", None)
        self.content.pop("tag", None)

        if details:
            self.content["issue"]["details"] = details
        if diagnostic:
            self.content["issue"]["diagnostic"] = diagnostic



import os
import json
from copy import deepcopy


# Do not resolve the following resources when parsing the schema.
definitions_blacklist = ["ResourceList"]  # ResourceList embeds all other resources,


def compatibilize_schema(resource):
    """
    As MongoDB only supports draft #4 of hte JSON schema standard,
    (https://tools.ietf.org/html/draft-zyp-json-schema-04)
    we have to manually migrate some of the keywords:
        - const
    """
    if "const" in resource:
        resource["enum"] = [resource["const"]]
        del resource["const"]


def enforce_id(schema):
    """
    As MongoDB automatically adds an '_id' property when insertign a document
    in the database, we have to update the schema accordingly.
    """
    schema["properties"]["_id"] = {"bsonType": "objectId"}


class SchemaParser:
    def __init__(self):
        """
        Initialized the parsed using the "fhir.schema.json" file
        located in the same directory.
        """
        schema_path = os.path.join(os.path.dirname(__file__), "fhir.schema.json")
        with open(schema_path, "r") as schemaFile:
            self.schema = json.load(schemaFile)
            self.definitions = self.schema["definitions"]
            self.resources = [os.path.basename(r["$ref"]) for r in self.schema["oneOf"]]

    def parse(self, resource=None, depth=4):
        """
        Parse the whole JSON schema or a single definition if `resource`
        is specified.

        Args:
            resource (optional): name of the resource (eg: 'Patient')
            depth (optional): Number of recursion to apply.

        Returns: a generator that yields each parsed schema as a tuple:
                    (resource_name, schema)
        """
        if resource is None:
            for resource in self.resources:
                schema = self.parse_schema(resource, depth)
                yield (resource, schema)
        else:
            schema = self.parse_schema(resource, depth)
            yield (resource, schema)

    def parse_schema(self, resource: str, depth: int):
        """
        Parses the JSON schema of a single definition.
        It also adds (or overrides) the property '_id' of the resource;
        as required by MongoDB.

        Args:
            resource: The name of the resource (eg: 'Patient')
            depth: Number of recursion to apply.

        Returns: The dereferenced schema.
                 (ready to be passed as mongoDB 'validator')
        """

        r = self.definitions[resource]
        dereferenced_schema = self.resolve(r, depth)
        enforce_id(dereferenced_schema)
        return dereferenced_schema

    def resolve(self, resource: dict, depth: int):
        """
        Dereference a JSON schema recursively.

        Args:
            resource: A single JSON schema definition.
                      Used in the recursions as either:
                        - an internal reference (dict with "$ref" key)
                        - a JSON schema object (dict with "properties" key)
                        - a JSON schema array (dict with "items" key)
                        - a JSON schema "oneOf" object (dict with "oneOf" key)
            depth: The number of internal references to recursively resolve.

        Returns: The dereferenced schema.

        """
        # stop resolving references if maximal depth is reached
        # the reference is removed from the resource but all other
        # properties are kept.
        if depth == 0 and "$ref" in resource:
            del resource["$ref"]
            return resource

        # resource is a reference
        # resolve its definition if not blacklisted.
        # (NB: deepcopy is required, resource will be modified in place)
        elif "$ref" in resource:
            path = os.path.basename(resource["$ref"])
            if path not in definitions_blacklist:
                return self.resolve(deepcopy(self.definitions[path]), depth - 1)
            del resource["$ref"]

        # resource is an object
        # resolve each of its properties recursively
        elif "properties" in resource:
            for k in list(resource["properties"].keys()):
                # only one level of recursion is allowed for extensions
                if k == "extension" or k == "modifierExtension":
                    resource["properties"][k] = self.resolve(
                        resource["properties"][k], depth if depth < 1 else 1
                    )
                else:
                    resource["properties"][k] = self.resolve(resource["properties"][k], depth)

        # resource is an array
        # resolve its "items" property
        elif "items" in resource:
            resource["items"] = self.resolve(resource["items"], depth)

        # resource is a "oneOf" object
        # resolve each of its "oneOf" properties
        elif "oneOf" in resource:
            resource["oneOf"] = [self.resolve(x, depth) for x in resource["oneOf"]]

        compatibilize_schema(resource)
        return resource
